#include "KXL.h"
#include <stdio.h>

/* macro for a safe call to DirectFB functions */
#define DFBCHECK(x...)                                                     \
{                                                                          \
               DFBResult err = x;                                          \
               if (err != DFB_OK) {                                        \
                    fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ ); \
                    DirectFBErrorFatal( #x, err );                         \
               }                                                           \
}

typedef struct {
  IDirectFBSurface *surface;
  IDirectFBSurface *sub_surface;
  Uint16            Width;       /**                **/
  Uint16            Height;      /** ⤵             **/
} KXL_Frame;

typedef struct {
  IDirectFB     *dfb;
  IDirectFBInputDevice *keyboard;
  IDirectFBInputBuffer *keybuffer;
  IDirectFBFont *font;

  DFBInputEvent  evt;

  Uint16         Depth;    /** ץ졼       **/
  Uint16         Width;    /**                **/
  Uint16         Height;   /** ⤵             **/
  KXL_Frame     *Frame;    /** ե졼Хåե **/

  Uint8          font_r;
  Uint8          font_g;
  Uint8          font_b;
} KXL_Window;

KXL_Window *KXL_Root;
Uint8       KXL_DName[65] = ""; /** ǥץ쥤̾   **/

struct {
  int width;
  int height;
} requested_mode = { 0, 0 }, best_mode = { 0, 0 };


int video_mode_callback( unsigned int  width,
			 unsigned int  height,
			 unsigned int  bpp,
			 void         *ctx );

/*==============================================================
   ǥץ쥤̾
     : ǥץ쥤̾
==============================================================*/
void KXL_DisplayName(Uint8 *name)
{
  sprintf(KXL_DName, "%s", name);
}

/*==============================================================
   եȥ顼
     : ɥΥݥ
==============================================================*/
void KXL_Font(Uint8 *str, Uint8 r, Uint8 g, Uint8 b)
{
  /* FIXME: parse X font description and load the font */

  KXL_Root->font_r = r;
  KXL_Root->font_g = g;
  KXL_Root->font_b = b;
}

/*==============================================================
    ̾   : ɥ
     : ɤ
           : ɥι⤵
           : ȥСɽʸ
           : X٥
     : ɥΥݥ 
==============================================================*/
void KXL_CreateWindow(Uint16 w, Uint16 h, Uint8 *title, Uint32 event)
{
  DFBResult ret;
  KXL_Rect rect;

  KXL_Root = (KXL_Window *)KXL_Malloc(sizeof(KXL_Window));

  DFBCHECK( DirectFBInit (NULL, NULL) );
  DFBCHECK( DirectFBCreate (&KXL_Root->dfb) );

  DFBCHECK( KXL_Root->dfb->CreateFont (KXL_Root->dfb, NULL, NULL, &KXL_Root->font) );

  DFBCHECK( KXL_Root->dfb->GetInputDevice (KXL_Root->dfb, DIDID_KEYBOARD , &KXL_Root->keyboard) );
  DFBCHECK( KXL_Root->keyboard->CreateInputBuffer (KXL_Root->keyboard, &KXL_Root->keybuffer) );

  DFBCHECK( KXL_Root->dfb->SetCooperativeLevel (KXL_Root->dfb, DFSCL_FULLSCREEN) );


  best_mode.width = best_mode.height = 0;

  requested_mode.width = w;
  requested_mode.height = h;

  fprintf( stderr, "Checking for best suiting mode for %dx%d...\n", w, h );

  ret = KXL_Root->dfb->EnumVideoModes( KXL_Root->dfb, video_mode_callback, NULL );
  if (ret)
    DirectFBError( "Could not enumerate video modes", ret );

  if (best_mode.width && best_mode.height)
    {
      fprintf( stderr, "Switching to mode %dx%d...\n", best_mode.width, best_mode.height );
      ret = KXL_Root->dfb->SetVideoMode(KXL_Root->dfb, best_mode.width, best_mode.height, 16);
      if (ret)
	DirectFBError( "Could not set resolution", ret );
    }

  KXL_Root->evt.type = 0;
  KXL_Root->Frame    = NULL;
  KXL_Root->Width    = w;
  KXL_Root->Height   = h;

  rect.Left   = 0;
  rect.Top    = 0;
  rect.Width  = w;
  rect.Height = h;
}

/*==============================================================
   ɥ
     : ɥΥݥ
==============================================================*/
void KXL_DeleteWindow(void)
{
  if (KXL_Root->Frame)
    {
      KXL_Root->Frame->sub_surface->Release( KXL_Root->Frame->sub_surface );
      KXL_Root->Frame->surface->Release( KXL_Root->Frame->surface );
      free(KXL_Root->Frame);
    }

  KXL_Root->font->Release( KXL_Root->font );
  KXL_Root->dfb->Release( KXL_Root->dfb );

  free(KXL_Root);
}

/*==============================================================
    ե졼
     : ɥΥݥ
           : ԥåޥåפβ
           : ԥåޥåפι⤵
     : ԥåޥåפΥݥ 
==============================================================*/
void KXL_ReSizeFrame(Uint16 w, Uint16 h)
{
  DFBResult ret;
  DFBSurfaceDescription dsc;
  int width, height;
  DFBRectangle rect;

  if (!KXL_Root)
    return;

  fprintf( stderr, "KXL_ReSizeFrame: %dx%d\n", w, h );

  if (KXL_Root->Frame)
    {
      KXL_Root->Frame->sub_surface->Release( KXL_Root->Frame->sub_surface );
      KXL_Root->Frame->surface->Release( KXL_Root->Frame->surface );
      free(KXL_Root->Frame);
    }

  KXL_Root->Frame         = (KXL_Frame *)KXL_Malloc(sizeof(KXL_Frame));

  dsc.flags = DSDESC_CAPS;
  dsc.caps  = DSCAPS_PRIMARY | DSCAPS_FLIPPING;

  DFBCHECK( KXL_Root->dfb->CreateSurface (KXL_Root->dfb, &dsc, &KXL_Root->Frame->surface) );

  KXL_Root->Frame->surface->GetSize( KXL_Root->Frame->surface,
				     &width, &height );
      
  rect.x = (width - w) / 2;
  rect.y = (height - h) / 2;
  rect.w = w;
  rect.h = h;

  DFBCHECK( KXL_Root->Frame->surface->GetSubSurface (KXL_Root->Frame->surface, &rect,
						     &KXL_Root->Frame->sub_surface) );

  KXL_Root->Frame->surface->SetColor (KXL_Root->Frame->surface, 0, 0, 0, 0);
  KXL_Root->Frame->surface->FillRectangle (KXL_Root->Frame->surface, 0, 0, width, height );
  KXL_Root->Frame->surface->Flip (KXL_Root->Frame->surface, NULL, 0 );
  KXL_Root->Frame->surface->FillRectangle (KXL_Root->Frame->surface, 0, 0, width, height );

  KXL_Root->Frame->sub_surface->SetBlittingFlags( KXL_Root->Frame->sub_surface,
						  DSBLIT_SRC_COLORKEY );

  KXL_Root->Frame->sub_surface->SetFont( KXL_Root->Frame->sub_surface, KXL_Root->font );

  KXL_Root->Frame->Width  = w;
  KXL_Root->Frame->Height = h;
}

/*==============================================================
    åץǡ
     : ɥΥݥ
           : ԥåޥåפοʿϰ
           : ԥåޥåפοľϰ
           : ԥåޥåפ
           : ԥåޥåפι⤵
==============================================================*/
void KXL_UpDate(KXL_Rect r)
{
  DFBRegion region = { r.Left, r.Top,
		       r.Left + r.Width - 1,
		       r.Top + r.Height - 1 };

  //  fprintf( stderr, "KXL_UpDate: %d, %d - %d, %d\n", region.x1, region.y1, region.x2, region.y2 );

  /*  if (r.Left   == 0  &&
      r.Top    == 0  &&
      r.Width  == KXL_Root->Frame->Width  &&
      r.Height == KXL_Root->Frame->Height)*/
    KXL_Root->Frame->surface->Flip( KXL_Root->Frame->surface, NULL, DSFLIP_WAITFORSYNC );
    /*  else
	KXL_Root->Frame->sub_surface->Flip( KXL_Root->Frame->sub_surface, &region, DSFLIP_WAITFORSYNC );*/
}

/*==============================================================
    ᡼ԡ
     : ɥΥݥ
           : ᡼Υݥ
           : ᡼οʿϰ
           : ᡼οľϰ
           : ᡼
           : ᡼ι⤵
==============================================================*/
KXL_Image *KXL_CopyImage(KXL_Image *src, KXL_Rect r)
{
  KXL_Image *dest;
  DFBRectangle rect = { r.Left, r.Top, r.Width, r.Height };
  DFBSurfacePixelFormat format;
  DFBSurfaceDescription dsc;

  src->surface->GetPixelFormat( src->surface, &format );

  dest         = (KXL_Image *)KXL_Malloc(sizeof(KXL_Image));
  dest->Width  = r.Width;
  dest->Height = r.Height;

  dsc.flags  = DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_PIXELFORMAT;
  dsc.width  = r.Width;
  dsc.height = r.Height;
  dsc.pixelformat = format;

  DFBCHECK( KXL_Root->dfb->CreateSurface (KXL_Root->dfb, &dsc, &dest->surface) );

  dest->surface->Blit( dest->surface, src->surface, &rect, 0, 0 );

  return dest;
}

/*==============================================================
    Υ᡼̾᡼򥳥ԡ
     : ɥΥݥ
           : ᡼Υݥ
           : ᡼
           : ᡼ι⤵
==============================================================*/
KXL_Image *KXL_StrechImage(KXL_Image *src, Uint16 width, Uint16 height)
{
  KXL_Image *dest;
  DFBSurfacePixelFormat format;
  DFBSurfaceDescription dsc;

  src->surface->GetPixelFormat( src->surface, &format );

  dest         = (KXL_Image *)KXL_Malloc(sizeof(KXL_Image));
  dest->Width  = width;
  dest->Height = height;

  dsc.flags  = DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_PIXELFORMAT;
  dsc.width  = width;
  dsc.height = height;
  dsc.pixelformat = format;

  DFBCHECK( KXL_Root->dfb->CreateSurface (KXL_Root->dfb, &dsc, &dest->surface) );

  dest->surface->StretchBlit( dest->surface, src->surface, NULL, NULL );

  return dest;
}

/*==============================================================
    ᡼
     : ɥΥݥ
           : ᡼Υݥ
==============================================================*/
void KXL_DeleteImage(KXL_Image *img)
{
  img->surface->Release( img->surface );

  free(img);
}

/*==============================================================
    XPMեɤ߹
     : ɥΥݥ
           : ե̾
    ͡᡼Υݥ
==============================================================*/
KXL_Image *KXL_LoadImage(Uint8 *filename)
{
  KXL_Image *img;
  int width, height;
  DFBSurfaceDescription dsc;
  IDirectFBImageProvider *provider;

  img = (KXL_Image *)KXL_Malloc(sizeof(KXL_Image));

  DFBCHECK( KXL_Root->dfb->CreateImageProvider (KXL_Root->dfb,
						filename, &provider) );

  DFBCHECK( provider->GetSurfaceDescription (provider, &dsc) );

  DFBCHECK( KXL_Root->dfb->CreateSurface (KXL_Root->dfb, &dsc, &img->surface) );

  DFBCHECK( provider->RenderTo (provider, img->surface) );

  img->surface->GetSize (img->surface, &width, &height);
  img->Width  = width;
  img->Height = height;

  provider->Release( provider );

  return img;
}

/*==============================================================
    ᡼
     : ɥΥݥ 
           : ᡼Υݥ
           : ʿϰ
           : ľϰ
==============================================================*/
void KXL_PutImage(KXL_Image *img, Sint16 x, Sint16 y)
{
  KXL_Root->Frame->sub_surface->Blit( KXL_Root->Frame->sub_surface,
				      img->surface, NULL, x, y );
}

/*==============================================================
    ᡼
     : ɥΥݥ 
           : ᡼Υݥ
           : ʿϰ
           : ľϰ
==============================================================*/
void KXL_PutRectImage(KXL_Image *img, KXL_Rect r, Sint16 x, Sint16 y)
{
  DFBRectangle rect = { r.Left, r.Top, r.Width, r.Height };
  
  KXL_Root->Frame->sub_surface->Blit( KXL_Root->Frame->sub_surface,
				      img->surface, &rect, x, y );
}

/*==============================================================
    ƥ
     : ɥΥݥ 
           : ʿϰ
           : ľϰ
	   : ʸΥݥ
==============================================================*/
void KXL_PutText(Sint16 x, Sint16 y, Uint8 *str)
{
  KXL_Root->Frame->sub_surface->SetColor( KXL_Root->Frame->sub_surface,
					  KXL_Root->font_r, KXL_Root->font_g,
					  KXL_Root->font_b, 0xff );

  KXL_Root->Frame->sub_surface->DrawString( KXL_Root->Frame->sub_surface,
					    str, -1, x, y, DSTF_LEFT );
}

/*==============================================================
    ƥ
     : ɥΥݥ 
           : ʿϰ
           : ľϰ
	   : ʸΥݥ
==============================================================*/
Uint16 KXL_TextWidth(Uint8 *str)
{
  int width;

  KXL_Root->font->GetStringWidth( KXL_Root->font, str, -1, &width );

  return width;
}

/*==============================================================
    ե졼९ꥢ
     : ɥΥݥ 
           : ʿϰ
           : ľϰ
	   : ʿ
	   ľ
==============================================================*/
void KXL_ClearFrame(KXL_Rect r)
{
  KXL_Root->Frame->sub_surface->SetColor( KXL_Root->Frame->sub_surface, 0, 0, 0, 0 );

  KXL_Root->Frame->sub_surface->FillRectangle( KXL_Root->Frame->sub_surface,
					       r.Left, r.Top, r.Width, r.Height );
}

/*==============================================================
    ɥ٥ȼ
     : ɥΥݥ 
    ͡٥ȥ
==============================================================*/
Uint32 KXL_GetEvents(void)
{
  switch (KXL_Root->evt.type)
    {
    case DIET_KEYPRESS:
      return KXL_EVENT_KEY_PRESS;
    case DIET_KEYRELEASE:
      return KXL_EVENT_KEY_RELEASE;
    case DIET_BUTTONPRESS:
      return KXL_EVENT_BUTTON_PRESS;
    case DIET_BUTTONRELEASE:
      return KXL_EVENT_BUTTON_RELEASE;
    }

  return 0;
}

/*==============================================================
    ɥ٥ȼ
     : ɥΥݥ 
    ͡٥ȥ
==============================================================*/
Bool KXL_CheckEvents(void)
{
  return (KXL_Root->keybuffer->GetEvent (KXL_Root->keybuffer, &KXL_Root->evt) == DFB_OK);
}

/*==============================================================
    
     : ɥΥݥ 
    ͡
==============================================================*/
Uint16 KXL_GetKey(void)
{
  switch (KXL_Root->evt.keycode)
    {
    case DIKC_A:   return KXL_KEY_a;
    case DIKC_B:   return KXL_KEY_b;
    case DIKC_C:   return KXL_KEY_c;
    case DIKC_D:   return KXL_KEY_d;
    case DIKC_E:   return KXL_KEY_e;
    case DIKC_F:   return KXL_KEY_f;
    case DIKC_G:   return KXL_KEY_g;
    case DIKC_H:   return KXL_KEY_h;
    case DIKC_I:   return KXL_KEY_i;
    case DIKC_J:   return KXL_KEY_j;
    case DIKC_K:   return KXL_KEY_k;
    case DIKC_L:   return KXL_KEY_l;
    case DIKC_M:   return KXL_KEY_m;
    case DIKC_N:   return KXL_KEY_n;
    case DIKC_O:   return KXL_KEY_o;
    case DIKC_P:   return KXL_KEY_p;
    case DIKC_Q:   return KXL_KEY_q;
    case DIKC_R:   return KXL_KEY_r;
    case DIKC_S:   return KXL_KEY_s;
    case DIKC_T:   return KXL_KEY_t;
    case DIKC_U:   return KXL_KEY_u;
    case DIKC_V:   return KXL_KEY_v;
    case DIKC_W:   return KXL_KEY_w;
    case DIKC_X:   return KXL_KEY_x;
    case DIKC_Y:   return KXL_KEY_y;
    case DIKC_Z:   return KXL_KEY_z;

    case DIKC_0:   return KXL_KEY_0;
    case DIKC_1:   return KXL_KEY_1;
    case DIKC_2:   return KXL_KEY_2;
    case DIKC_3:   return KXL_KEY_3;
    case DIKC_4:   return KXL_KEY_4;
    case DIKC_5:   return KXL_KEY_5;
    case DIKC_6:   return KXL_KEY_6;
    case DIKC_7:   return KXL_KEY_7;
    case DIKC_8:   return KXL_KEY_8;
    case DIKC_9:   return KXL_KEY_9;

    case DIKC_F1:  return KXL_KEY_F1;
    case DIKC_F2:  return KXL_KEY_F2;
    case DIKC_F3:  return KXL_KEY_F3;
    case DIKC_F4:  return KXL_KEY_F4;
    case DIKC_F5:  return KXL_KEY_F5;
    case DIKC_F6:  return KXL_KEY_F6;
    case DIKC_F7:  return KXL_KEY_F7;
    case DIKC_F8:  return KXL_KEY_F8;
    case DIKC_F9:  return KXL_KEY_F9;
    case DIKC_F10: return KXL_KEY_F10;
    case DIKC_F11: return KXL_KEY_F11;
    case DIKC_F12: return KXL_KEY_F12;

    case DIKC_ESCAPE:    return KXL_KEY_Escape;
    case DIKC_LEFT:      return KXL_KEY_Left;
    case DIKC_RIGHT:     return KXL_KEY_Right;
    case DIKC_UP:        return KXL_KEY_Up;
    case DIKC_DOWN:      return KXL_KEY_Down;
    case DIKC_CTRL:      return KXL_KEY_Control_L;
    case DIKC_SHIFT:     return KXL_KEY_Shift_L;
    case DIKC_TAB:       return KXL_KEY_Tab;
    case DIKC_ENTER:     return KXL_KEY_Return;
    case DIKC_SPACE:     return KXL_KEY_Space;
    case DIKC_BACKSPACE: return KXL_KEY_BackSpace;
    case DIKC_DELETE:    return KXL_KEY_Delete;
    case DIKC_HOME:      return KXL_KEY_Home;
    case DIKC_KP_DIV:    return KXL_KEY_Slash;
    case DIKC_KP_MULT:   return KXL_KEY_Asterisk;
    case DIKC_KP_MINUS:  return KXL_KEY_Minus;
    case DIKC_KP_PLUS:   return KXL_KEY_Plus;
    case DIKC_KP_ENTER:  return KXL_KEY_Return;

    default:
      break;
    }

  return 0;
}

int video_mode_callback( unsigned int  width,
			 unsigned int  height,
			 unsigned int  bpp,
			 void         *ctx )
{
  if (bpp == 16 && width >= requested_mode.width && height >= requested_mode.height)
    {
      if (best_mode.width == 0  ||  best_mode.width > width || best_mode.height > height)
	{
	  best_mode.width = width;
	  best_mode.height = height;
	}
    }
}
